@@ -14,14 +14,16 @@
 import concurrent
 import json
 import logging
+import multiprocessing
 import os
 import random
 import shutil
 import signal
+import sys
 import tempfile
 import traceback
-import types
 from abc import abstractmethod
+from contextlib import suppress
 from dataclasses import dataclass
 from multiprocessing import Process, Queue
 from pathlib import Path
@@ -30,43 +32,32 @@ from time import sleep, time
 from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union
 from urllib import parse
 
+import boto3
+import botocore
 import numpy as np
 import torch
 
 from litdata.constants import (
-    _BOTO3_AVAILABLE,
     _DEFAULT_FAST_DEV_RUN_ITEMS,
+    _ENABLE_STATUS,
     _INDEX_FILENAME,
     _IS_IN_STUDIO,
-    _LIGHTNING_CLOUD_AVAILABLE,
-    _TORCH_GREATER_EQUAL_2_1_0,
+    _TQDM_AVAILABLE,
 )
-from litdata.imports import RequirementCache
 from litdata.processing.readers import BaseReader, StreamingDataLoaderReader
-from litdata.processing.utilities import _create_dataset
+from litdata.processing.utilities import _create_dataset, download_directory_from_S3, remove_uuid_from_filename
 from litdata.streaming import Cache
 from litdata.streaming.cache import Dir
 from litdata.streaming.client import S3Client
 from litdata.streaming.dataloader import StreamingDataLoader
+from litdata.streaming.item_loader import BaseItemLoader
 from litdata.streaming.resolver import _resolve_dir
+from litdata.utilities._pytree import tree_flatten, tree_unflatten, treespec_loads
 from litdata.utilities.broadcast import broadcast_object
+from litdata.utilities.dataset_utilities import load_index_file
+from litdata.utilities.encryption import Encryption
 from litdata.utilities.packing import _pack_greedily
 
-_TQDM_AVAILABLE = RequirementCache("tqdm")
-
-if _TQDM_AVAILABLE:
-    from tqdm.auto import tqdm as _tqdm
-
-if _TORCH_GREATER_EQUAL_2_1_0:
-    from torch.utils._pytree import tree_flatten, tree_unflatten, treespec_loads
-
-if _LIGHTNING_CLOUD_AVAILABLE:
-    from lightning_cloud.openapi import V1DatasetType
-
-
-if _BOTO3_AVAILABLE:
-    import botocore
-
 logger = logging.Logger(__name__)
 
 
@@ -234,10 +225,16 @@ def _upload_fn(upload_queue: Queue, remove_queue: Queue, cache_dir: str, output_
 
         if obj.scheme == "s3":
             try:
+                output_filepath = str(obj.path).lstrip("/")
+
+                if local_filepath.__contains__(".checkpoints"):
+                    output_filepath = os.path.join(output_filepath, ".checkpoints")
                 if tmpdir is None:
-                    output_filepath = os.path.join(str(obj.path).lstrip("/"), os.path.basename(local_filepath))
+                    output_filepath = os.path.join(output_filepath, os.path.basename(local_filepath))
                 else:
-                    output_filepath = os.path.join(str(obj.path).lstrip("/"), local_filepath.replace(tmpdir, "")[1:])
+                    output_filepath = os.path.join(output_filepath, local_filepath.replace(tmpdir, "")[1:])
+
+                output_filepath = remove_uuid_from_filename(output_filepath)  # remove unique id from checkpoints
 
                 s3.client.upload_file(
                     local_filepath,
@@ -248,13 +245,20 @@ def _upload_fn(upload_queue: Queue, remove_queue: Queue, cache_dir: str, output_
                 print(e)
 
         elif output_dir.path:
+            output_filepath = output_dir.path
+
+            if local_filepath.__contains__(".checkpoints"):
+                output_filepath = os.path.join(output_filepath, ".checkpoints")
+
             if tmpdir is None:
-                output_filepath = os.path.join(output_dir.path, os.path.basename(local_filepath))
+                output_filepath = os.path.join(output_filepath, os.path.basename(local_filepath))
             else:
-                output_filepath = os.path.join(output_dir.path, local_filepath.replace(tmpdir, "")[1:])
+                output_filepath = os.path.join(output_filepath, local_filepath.replace(tmpdir, "")[1:])
+
+            output_filepath = remove_uuid_from_filename(output_filepath)  # remove unique id from checkpoints
 
             os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
-            shutil.move(local_filepath, output_filepath)
+            shutil.copy(local_filepath, output_filepath)
         else:
             raise ValueError(f"The provided {output_dir.path} isn't supported.")
 
@@ -375,6 +379,22 @@ def _is_path(input_dir: Optional[str], element: Any) -> bool:
     return os.path.isfile(element)
 
 
+class FakeQueue:
+    """This class enables us to replace multiprocessing Queue when not required and avoid serializing data."""
+
+    def __init__(self) -> None:
+        self._items: List[Any] = []
+
+    def add_items(self, items: List[Any]) -> None:
+        self._items.extend(items)
+
+    def get(self) -> None:
+        try:
+            return self._items.pop(0)
+        except IndexError:
+            return None
+
+
 class BaseWorker:
     def __init__(
         self,
@@ -392,6 +412,11 @@ class BaseWorker:
         num_uploaders: int,
         remove: bool,
         reader: Optional[BaseReader] = None,
+        writer_starting_chunk_index: int = 0,
+        use_checkpoint: bool = False,
+        checkpoint_chunks_info: Optional[List[Dict[str, Any]]] = None,
+        checkpoint_next_index: Optional[int] = None,
+        item_loader: Optional[BaseItemLoader] = None,
     ) -> None:
         """The BaseWorker is responsible to process the user data."""
         self.worker_index = worker_index
@@ -413,13 +438,19 @@ class BaseWorker:
         self.to_download_queues: List[Queue] = []
         self.to_upload_queues: List[Queue] = []
         self.stop_queue = stop_queue
-        self.ready_to_process_queue: Queue = Queue()
+        self.no_downloaders = self.input_dir.path is None or self.reader is not None
+        self.ready_to_process_queue: Union[Queue, FakeQueue] = FakeQueue() if self.no_downloaders else Queue()
         self.remove_queue: Queue = Queue()
         self.progress_queue: Queue = progress_queue
         self.error_queue: Queue = error_queue
+        self.item_loader = item_loader
         self._counter = 0
         self._last_time = time()
         self._index_counter = 0
+        self.writer_starting_chunk_index: int = writer_starting_chunk_index
+        self.use_checkpoint: bool = use_checkpoint
+        self.checkpoint_chunks_info: Optional[List[Dict[str, Any]]] = checkpoint_chunks_info
+        self.checkpoint_next_index: Optional[int] = checkpoint_next_index
 
     def run(self) -> None:
         try:
@@ -427,7 +458,6 @@ class BaseWorker:
             self._loop()
         except Exception:
             traceback_format = traceback.format_exc()
-            print(traceback_format)
             self.error_queue.put(traceback_format)
         print(f"Worker {str(_get_node_rank() * self.num_workers + self.worker_index)} is done.")
 
@@ -500,10 +530,7 @@ class BaseWorker:
 
     def _create_cache(self) -> None:
         self.cache_data_dir = _get_cache_data_dir()
-        os.makedirs(self.cache_data_dir, exist_ok=True)
-
         self.cache_chunks_dir = _get_cache_dir()
-        os.makedirs(self.cache_chunks_dir, exist_ok=True)
 
         if isinstance(self.data_recipe, DataTransformRecipe):
             return
@@ -513,9 +540,25 @@ class BaseWorker:
             chunk_bytes=self.data_recipe.chunk_bytes,
             chunk_size=self.data_recipe.chunk_size,
             compression=self.data_recipe.compression,
+            encryption=self.data_recipe.encryption,
+            writer_chunk_index=self.writer_starting_chunk_index,
+            item_loader=self.item_loader,
         )
         self.cache._reader._rank = _get_node_rank() * self.num_workers + self.worker_index
 
+        # return
+        if self.use_checkpoint and all(
+            [
+                self.checkpoint_chunks_info is not None,
+                self.checkpoint_next_index is not None,
+            ]
+        ):
+            assert isinstance(self.checkpoint_next_index, int)
+            assert isinstance(self.checkpoint_chunks_info, list)
+
+            self.cache._writer._chunks_info = self.checkpoint_chunks_info
+            self.cache._writer._chunk_index += self.checkpoint_next_index
+
     def _try_upload(self, data: Optional[Union[str, Tuple[str, str]]]) -> None:
         if not data or (self.output_dir.url if self.output_dir.url else self.output_dir.path) is None:
             return
@@ -528,11 +571,12 @@ class BaseWorker:
         self.to_upload_queues[self._counter % self.num_uploaders].put(data)
 
     def _collect_paths(self) -> None:
-        if self.input_dir.path is None or self.reader is not None:
-            for index in range(len(self.items)):
-                self.ready_to_process_queue.put(index)
-            for _ in range(self.num_downloaders):
-                self.ready_to_process_queue.put(None)
+        if self.no_downloaders:
+            if isinstance(self.ready_to_process_queue, FakeQueue):
+                self.ready_to_process_queue.add_items(list(range(len(self.items))))
+            else:
+                for index in range(len(self.items)):
+                    self.ready_to_process_queue.put(index)
             return
 
         items = []
@@ -556,7 +600,11 @@ class BaseWorker:
             paths = []
             for index, path in indexed_paths.items():
                 paths.append(path)
-                if self.input_dir and not self.input_dir.path.startswith("/teamspace/studios/this_studio"):
+                if (
+                    self.input_dir
+                    and isinstance(self.input_dir.path, str)
+                    and not self.input_dir.path.startswith("/teamspace/studios/this_studio")
+                ):
                     path = path.replace(self.input_dir.path, self.cache_data_dir)
                 flattened_item[index] = path
 
@@ -567,7 +615,7 @@ class BaseWorker:
         self.items = items
 
     def _start_downloaders(self) -> None:
-        if self.input_dir.path is None or self.reader is not None:
+        if self.no_downloaders:
             return
 
         for _ in range(self.num_downloaders):
@@ -628,7 +676,7 @@ class BaseWorker:
         try:
             current_item = self.items[index] if self.reader is None else self.reader.read(self.items[index])
             item_data_or_generator = self.data_recipe.prepare_item(current_item)
-            if isinstance(item_data_or_generator, types.GeneratorType):
+            if self.data_recipe.is_generator:
                 for item_data in item_data_or_generator:
                     if item_data is not None:
                         chunk_filepath = self.cache._add_item(self._index_counter, item_data)
@@ -638,8 +686,11 @@ class BaseWorker:
                 chunk_filepath = self.cache._add_item(self._index_counter, item_data_or_generator)
                 self._try_upload(chunk_filepath)
                 self._index_counter += 1
+                if self.use_checkpoint:
+                    checkpoint_filepath = self.cache.save_checkpoint()
+                    self._try_upload(checkpoint_filepath)
         except Exception as e:
-            raise RuntimeError(f"Failed processing {self.items[index]}") from e
+            raise RuntimeError(f"Failed processing {self.items[index]=}; {index=}") from e
 
     def _handle_data_chunk_recipe_end(self) -> None:
         chunks_filepaths = self.cache.done()
@@ -649,6 +700,10 @@ class BaseWorker:
                 if isinstance(chunk_filepath, str) and os.path.exists(chunk_filepath):
                     self.to_upload_queues[i % self.num_uploaders].put(chunk_filepath)
 
+        if self.use_checkpoint and not self.data_recipe.is_generator:
+            checkpoint_filepath = self.cache.save_checkpoint()
+            self._try_upload(checkpoint_filepath)
+
     def _handle_data_transform_recipe(self, index: int) -> None:
         # Don't use a context manager to avoid deleting files that are being uploaded.
         output_dir = tempfile.mkdtemp()
@@ -681,6 +736,7 @@ class _Result:
     num_bytes: Optional[str] = None
     data_format: Optional[str] = None
     compression: Optional[str] = None
+    encryption: Optional[Encryption] = None
     num_chunks: Optional[int] = None
     num_bytes_per_chunk: Optional[List[int]] = None
 
@@ -710,6 +766,7 @@ class DataChunkRecipe(DataRecipe):
         chunk_size: Optional[int] = None,
         chunk_bytes: Optional[Union[int, str]] = None,
         compression: Optional[str] = None,
+        encryption: Optional[Encryption] = None,
     ):
         super().__init__()
         if chunk_size is not None and chunk_bytes is not None:
@@ -718,6 +775,7 @@ class DataChunkRecipe(DataRecipe):
         self.chunk_size = chunk_size
         self.chunk_bytes = 1 << 26 if chunk_size is None and chunk_bytes is None else chunk_bytes
         self.compression = compression
+        self.encryption = encryption
 
     @abstractmethod
     def prepare_structure(self, input_dir: Optional[str]) -> List[T]:
@@ -741,12 +799,12 @@ class DataChunkRecipe(DataRecipe):
 
         merge_cache = Cache(cache_dir, chunk_bytes=1)
         node_rank = _get_node_rank()
-        merge_cache._merge_no_wait(node_rank if num_nodes > 1 else None)
+        merge_cache._merge_no_wait(node_rank if num_nodes > 1 else None, getattr(self, "existing_index", None))
+
         self._upload_index(output_dir, cache_dir, num_nodes, node_rank)
 
         if num_nodes == node_rank + 1:
-            with open(os.path.join(cache_dir, _INDEX_FILENAME)) as f:
-                config = json.load(f)
+            config = load_index_file(cache_dir)
 
             size = sum([c["dim"] if c["dim"] is not None else c["chunk_size"] for c in config["chunks"]])
             num_bytes = sum([c["chunk_bytes"] for c in config["chunks"]])
@@ -761,7 +819,6 @@ class DataChunkRecipe(DataRecipe):
             # The platform can't store more than 1024 entries.
             # Note: This isn't really used right now, so it is fine to skip if too big.
             num_bytes_per_chunk = [c["chunk_size"] for c in config["chunks"]] if num_chunks < 1024 else []
-
             return _Result(
                 size=size,
                 num_bytes=num_bytes,
@@ -847,6 +904,10 @@ class DataProcessor:
         reorder_files: bool = True,
         weights: Optional[List[int]] = None,
         reader: Optional[BaseReader] = None,
+        state_dict: Optional[Dict[int, int]] = None,
+        use_checkpoint: bool = False,
+        item_loader: Optional[BaseItemLoader] = None,
+        start_method: Optional[str] = None,
     ):
         """The `DatasetOptimiser` provides an efficient way to process data across multiple machine into chunks to make
         training faster.
@@ -865,14 +926,33 @@ class DataProcessor:
             weights: Provide a list of weights associated to the inputs.
                 This is used to evenly split the work among the workers.
             reader: Map the inputs to worker inputs and provides a read method to read a slice of the data.
+            state_dict: The writer state dict. This is used to decide how to append data to an existing dataset.
+            use_checkpoint: Whether to create checkpoints while processing the data, which can be used to resume the
+                processing from the last checkpoint if the process is interrupted. (`Default: False`)
+            item_loader: The item loader that will be used during loading in StreamingDataset. Determines
+                    the format in which the data is stored and optimized for loading.
+            start_method: The start method used by python multiprocessing package. Default to spawn unless running
+                inside an interactive shell like Ipython.
 
         """
+        # spawn doesn't work in IPython
+        start_method = start_method or ("fork" if in_notebook() else "spawn")
+
+        msg = f"Setting multiprocessing start_method to {start_method}. "
+        if in_notebook() and start_method == "fork":
+            msg += "Tip: Libraries relying on lock can hang with `fork`. To use `spawn` in notebooks, "
+            msg += "move your code to files and import it within the notebook."
+
+        print(msg)
+
+        multiprocessing.set_start_method(start_method, force=True)
+
         self.input_dir = _resolve_dir(input_dir)
         self.output_dir = _resolve_dir(output_dir)
 
         self.num_workers = num_workers or (1 if fast_dev_run else (os.cpu_count() or 1) * 4)
         self.num_downloaders = num_downloaders or 2
-        self.num_uploaders = num_uploaders or 5
+        self.num_uploaders = num_uploaders or 1
         self.delete_cached_files = delete_cached_files
         self.fast_dev_run = _get_fast_dev_run() if fast_dev_run is None else fast_dev_run
         self.workers: Any = []
@@ -883,12 +963,18 @@ class DataProcessor:
         self.reorder_files = reorder_files
         self.weights = weights
         self.reader = reader
+        self.use_checkpoint = use_checkpoint
+        self.checkpoint_chunks_info: Optional[List[List[Dict[str, Any]]]] = None
+        self.checkpoint_next_index: Optional[List[int]] = None
+        self.item_loader = item_loader
+
+        self.state_dict = state_dict or {rank: 0 for rank in range(self.num_workers)}
 
         if self.reader is not None and self.weights is not None:
             raise ValueError("Either the reader or the weights needs to be defined.")
 
         # Ensure the input dir is the same across all nodes
-        self.input_dir = broadcast_object("input_dir", self.input_dir)
+        self.input_dir = broadcast_object("input_dir", self.input_dir, rank=_get_node_rank())
 
         if self.output_dir:
             # Ensure the output dir is the same across all nodes
@@ -901,6 +987,9 @@ class DataProcessor:
         """The `DataProcessor.run(...)` method triggers the data recipe processing over your dataset."""
         if not isinstance(data_recipe, DataRecipe):
             raise ValueError("The provided value should be a data recipe.")
+        if not self.use_checkpoint and isinstance(data_recipe, DataChunkRecipe):
+            # clean up checkpoints if not using checkpoints
+            self._cleanup_checkpoints()
 
         t0 = time()
         print(f"Setup started with fast_dev_run={self.fast_dev_run}.")
@@ -911,8 +1000,9 @@ class DataProcessor:
         torch.manual_seed(self.random_seed)
 
         # Call the setup method of the user
-        user_items: List[Any] = data_recipe.prepare_structure(self.input_dir.path if self.input_dir else None)
-
+        user_items: Union[List[Any], StreamingDataLoader] = data_recipe.prepare_structure(
+            self.input_dir.path if self.input_dir else None
+        )
         if not isinstance(user_items, (list, StreamingDataLoader)):
             raise ValueError("The `prepare_structure` should return a list of item metadata.")
 
@@ -922,6 +1012,8 @@ class DataProcessor:
         if self.reader:
             user_items = self.reader.remap_items(user_items, self.num_workers)
 
+        assert isinstance(user_items, list)
+
         if self.weights is not None:
             if len(self.weights) != len(user_items):
                 raise ValueError("The provided weights length should match the inputs' length.")
@@ -940,8 +1032,28 @@ class DataProcessor:
 
         print(f"Setup finished in {round(time() - t0, 3)} seconds. Found {len(user_items)} items to process.")
 
+        if self.use_checkpoint:
+            if hasattr(data_recipe, "is_generator") and data_recipe.is_generator:
+                # Checkpoint feature is not supported for generators for now.
+                raise ValueError("Checkpoint feature is not supported for generators, yet.")
+            # get the last checkpoint details
+            print("Resuming from last saved checkpoint...")
+            self._load_checkpoint_config(workers_user_items)
+
+            assert isinstance(self.checkpoint_next_index, list)
+
+            if all(self.checkpoint_next_index[i] == 0 for i in range(self.num_workers)):
+                # save the current configuration in the checkpoints.json file
+                print("No checkpoints found. Saving current configuration...")
+                self._save_current_config(workers_user_items)
+            else:
+                # load the last checkpoint details
+                assert isinstance(self.checkpoint_next_index, list)
+                workers_user_items = [w[self.checkpoint_next_index[i] :] for i, w in enumerate(workers_user_items)]
+                print("Checkpoints loaded successfully.")
+
         if self.fast_dev_run:
-            items_to_keep = self.fast_dev_run if type(self.fast_dev_run) is int else _DEFAULT_FAST_DEV_RUN_ITEMS
+            items_to_keep = self.fast_dev_run if isinstance(self.fast_dev_run, int) else _DEFAULT_FAST_DEV_RUN_ITEMS
             workers_user_items = [w[:items_to_keep] for w in workers_user_items]
             print(f"Fast dev run is enabled. Limiting to {items_to_keep} items per process.")
 
@@ -966,6 +1078,8 @@ class DataProcessor:
 
         current_total = 0
         if _TQDM_AVAILABLE:
+            from tqdm.auto import tqdm as _tqdm
+
             pbar = _tqdm(
                 desc="Progress",
                 total=num_items,
@@ -999,7 +1113,7 @@ class DataProcessor:
             if current_total == num_items:
                 break
 
-            if _IS_IN_STUDIO and node_rank == 0:
+            if _IS_IN_STUDIO and node_rank == 0 and _ENABLE_STATUS:
                 with open("status.json", "w") as f:
                     json.dump({"progress": str(100 * current_total * num_nodes / total_num_items) + "%"}, f)
 
@@ -1011,16 +1125,17 @@ class DataProcessor:
         if _TQDM_AVAILABLE:
             pbar.close()
 
-        # TODO: Understand why it hangs.
+        # TODO: Check whether this is still required.
         if num_nodes == 1:
             for w in self.workers:
-                w.join(0)
+                w.join()
 
         print("Workers are finished.")
         result = data_recipe._done(len(user_items), self.delete_cached_files, self.output_dir)
 
-        if num_nodes == node_rank + 1 and self.output_dir.url and _IS_IN_STUDIO and self.input_dir.path:
-            assert self.output_dir.path
+        if num_nodes == node_rank + 1 and self.output_dir.url and self.output_dir.path is not None and _IS_IN_STUDIO:
+            from lightning_cloud.openapi import V1DatasetType
+
             _create_dataset(
                 input_dir=self.input_dir.path,
                 storage_dir=self.output_dir.path,
@@ -1037,10 +1152,14 @@ class DataProcessor:
             )
 
         print("Finished data processing!")
+        if self.use_checkpoint and isinstance(data_recipe, DataChunkRecipe):
+            # clean up checkpoints
+            self._cleanup_checkpoints()
 
     def _exit_on_error(self, error: str) -> None:
         for w in self.workers:
-            w.join(0)
+            # w.join(0)
+            w.terminate()  # already error has occurred. So, no benefit of processing further.
         raise RuntimeError(f"We found the following error {error}.")
 
     def _create_process_workers(self, data_recipe: DataRecipe, workers_user_items: List[List[Any]]) -> None:
@@ -1064,6 +1183,11 @@ class DataProcessor:
                 self.num_uploaders,
                 self.delete_cached_files,
                 self.reader,
+                self.state_dict[worker_idx],
+                self.use_checkpoint,
+                self.checkpoint_chunks_info[worker_idx] if self.checkpoint_chunks_info else None,
+                self.checkpoint_next_index[worker_idx] if self.checkpoint_next_index else None,
+                self.item_loader,
             )
             worker.start()
             workers.append(worker)
@@ -1096,3 +1220,179 @@ class DataProcessor:
             shutil.rmtree(cache_data_dir, ignore_errors=True)
 
         os.makedirs(cache_data_dir, exist_ok=True)
+
+    def _cleanup_checkpoints(self) -> None:
+        if not isinstance(self.output_dir, Dir):
+            raise ValueError("The provided output_dir isn't a Dir Object.")
+
+        if self.output_dir.url is None:
+            # this is a local directory
+            if self.output_dir.path is None:
+                return
+
+            if os.path.exists(self.output_dir.path):
+                # clear the checkpoints
+                with suppress(FileNotFoundError):
+                    shutil.rmtree(os.path.join(self.output_dir.path, ".checkpoints"))
+
+            return
+
+        obj = parse.urlparse(self.output_dir.url)
+
+        if obj.scheme != "s3":
+            raise ValueError(f"The provided folder should start with s3://. Found {self.output_dir.path}.")
+
+        s3 = boto3.client("s3")
+
+        prefix = obj.path.lstrip("/").rstrip("/") + "/"
+
+        # Delete all the files (including the index file in overwrite mode)
+        bucket_name = obj.netloc
+        s3 = boto3.resource("s3")
+
+        checkpoint_prefix = os.path.join(prefix, ".checkpoints")
+
+        for obj in s3.Bucket(bucket_name).objects.filter(Prefix=checkpoint_prefix):
+            s3.Object(bucket_name, obj.key).delete()
+
+    def _save_current_config(self, workers_user_items: List[List[Any]]) -> None:
+        if not self.use_checkpoint:
+            return
+
+        # save the current configuration in the config.json file
+        config = {
+            "num_workers": self.num_workers,
+            "workers_user_items": workers_user_items,
+        }
+
+        try:
+            if self.output_dir.url is None:
+                assert self.output_dir.path
+
+                if not os.path.exists(os.path.join(self.output_dir.path, ".checkpoints")):
+                    os.makedirs(os.path.join(self.output_dir.path, ".checkpoints"))
+
+                with open(os.path.join(self.output_dir.path, ".checkpoints", "config.json"), "w") as f:
+                    json.dump(config, f)
+
+                return
+
+            obj = parse.urlparse(self.output_dir.url)
+
+            if obj.scheme != "s3":
+                raise ValueError(f"The provided folder should start with s3://. Found {self.output_dir.path}.")
+
+            # TODO: Add support for all cloud providers
+
+            s3 = S3Client()
+
+            prefix = obj.path.lstrip("/").rstrip("/") + "/" + ".checkpoints/"
+
+            # write config.json file to temp directory and upload it to s3
+            with tempfile.TemporaryDirectory() as temp_dir:
+                temp_file_name = os.path.join(temp_dir, "config.json")
+                with open(temp_file_name, "w") as f:
+                    json.dump(config, f)
+                s3.client.upload_file(
+                    temp_file_name,
+                    obj.netloc,
+                    os.path.join(prefix, "config.json"),
+                )
+        except Exception as e:
+            print(e)
+
+    def _load_checkpoint_config(self, workers_user_items: List[List[Any]]) -> None:
+        if not self.use_checkpoint:
+            return
+
+        default_chunk_info: List[Dict[str, Any]] = []
+
+        self.checkpoint_chunks_info = [default_chunk_info for _ in range(self.num_workers)]
+        self.checkpoint_next_index = [0 for _ in range(self.num_workers)]
+
+        if self.output_dir.url is None:
+            assert self.output_dir.path
+
+            if not os.path.exists(os.path.join(self.output_dir.path, ".checkpoints")):
+                return
+
+            if not os.path.exists(os.path.join(self.output_dir.path, ".checkpoints", "config.json")):
+                # if the config.json file doesn't exist, we don't have any checkpoint saved
+                return
+
+            with open(os.path.join(self.output_dir.path, ".checkpoints", "config.json")) as f:
+                config = json.load(f)
+
+            if config["num_workers"] != self.num_workers:
+                raise ValueError(
+                    "The number of workers in the checkpoints doesn't match the current number of workers."
+                )
+
+            if config["workers_user_items"] != workers_user_items:
+                raise ValueError("Existing checkpoints are not compatible with the current configuration.")
+
+            checkpoint_file_names = [f"checkpoint-{worker_idx}.json" for worker_idx in range(self.num_workers)]
+
+            for i, checkpoint_file_name in enumerate(checkpoint_file_names):
+                if not os.path.exists(os.path.join(self.output_dir.path, ".checkpoints", checkpoint_file_name)):
+                    # if the checkpoint file doesn't exist, we don't have any checkpoint saved for this worker
+                    continue
+
+                with open(os.path.join(self.output_dir.path, ".checkpoints", checkpoint_file_name)) as f:
+                    checkpoint = json.load(f)
+
+                self.checkpoint_chunks_info[i] = checkpoint["chunks"]
+                self.checkpoint_next_index[i] = checkpoint["done_till_index"]
+            return
+
+        obj = parse.urlparse(self.output_dir.url)
+
+        if obj.scheme != "s3":
+            raise ValueError(f"The provided folder should start with s3://. Found {self.output_dir.path}.")
+
+        # TODO: Add support for all cloud providers
+
+        prefix = obj.path.lstrip("/").rstrip("/") + "/" + ".checkpoints/"
+
+        # Delete all the files (including the index file in overwrite mode)
+        bucket_name = obj.netloc
+
+        # download all the checkpoint files in tempdir and read them
+        with tempfile.TemporaryDirectory() as temp_dir:
+            saved_file_dir = download_directory_from_S3(bucket_name, prefix, temp_dir)
+
+            if not os.path.exists(os.path.join(saved_file_dir, "config.json")):
+                # if the config.json file doesn't exist, we don't have any checkpoint saved
+                return
+
+            # read the config.json file
+            with open(os.path.join(saved_file_dir, "config.json")) as f:
+                config = json.load(f)
+
+            if config["num_workers"] != self.num_workers:
+                raise ValueError(
+                    "The number of workers in the checkpoints doesn't match the current number of workers."
+                )
+
+            if config["workers_user_items"] != workers_user_items:
+                raise ValueError("Existing checkpoints are not compatible with the current configuration.")
+
+            checkpoint_file_names = [f"checkpoint-{worker_idx}.json" for worker_idx in range(self.num_workers)]
+
+            for i, checkpoint_file_name in enumerate(checkpoint_file_names):
+                if not os.path.exists(os.path.join(saved_file_dir, checkpoint_file_name)):
+                    # if the checkpoint file doesn't exist, we don't have any checkpoint saved for this worker
+                    continue
+
+                with open(os.path.join(saved_file_dir, checkpoint_file_name)) as f:
+                    checkpoint = json.load(f)
+
+                self.checkpoint_chunks_info[i] = checkpoint["chunks"]
+                self.checkpoint_next_index[i] = checkpoint["done_till_index"]
+        return
+
+
+def in_notebook() -> bool:
+    """Returns ``True`` if the module is running in IPython kernel, ``False`` if in IPython shell or other Python
+    shell."""
+    return "ipykernel" in sys.modules
