@@ -13,20 +13,32 @@
 
 import concurrent.futures
 import inspect
+import json
 import os
+import shutil
+import tempfile
+from dataclasses import dataclass
 from datetime import datetime
 from functools import partial
 from pathlib import Path
 from types import FunctionType
-from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
+from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union
+from urllib import parse
 
 import torch
 
-from litdata.constants import _IS_IN_STUDIO, _TORCH_GREATER_EQUAL_2_1_0
+from litdata.constants import _INDEX_FILENAME, _IS_IN_STUDIO
 from litdata.processing.data_processor import DataChunkRecipe, DataProcessor, DataTransformRecipe
 from litdata.processing.readers import BaseReader
-from litdata.processing.utilities import optimize_dns_context
+from litdata.processing.utilities import (
+    _get_work_dir,
+    extract_rank_and_index_from_filename,
+    optimize_dns_context,
+    read_index_file_content,
+)
+from litdata.streaming.client import S3Client
 from litdata.streaming.dataloader import StreamingDataLoader
+from litdata.streaming.item_loader import BaseItemLoader
 from litdata.streaming.resolver import (
     Dir,
     _assert_dir_has_index_file,
@@ -34,9 +46,14 @@ from litdata.streaming.resolver import (
     _execute,
     _resolve_dir,
 )
+from litdata.utilities._pytree import tree_flatten
+from litdata.utilities.encryption import Encryption
+from litdata.utilities.format import _get_tqdm_iterator_if_available
 
-if _TORCH_GREATER_EQUAL_2_1_0:
-    from torch.utils._pytree import tree_flatten
+
+def _is_remote_file(path: str) -> bool:
+    obj = parse.urlparse(path)
+    return obj.scheme in ["s3", "gcs"]
 
 
 def _get_indexed_paths(data: Any) -> Dict[int, str]:
@@ -45,7 +62,7 @@ def _get_indexed_paths(data: Any) -> Dict[int, str]:
     return {
         index: element
         for index, element in enumerate(flattened_item)
-        if isinstance(element, str) and os.path.exists(element)
+        if isinstance(element, str) and (os.path.exists(element) or _is_remote_file(element))
     }
 
 
@@ -63,12 +80,16 @@ def _get_input_dir(inputs: Sequence[Any]) -> Optional[str]:
         # Every element should have filepaths if any contains one.
         raise ValueError(f"The provided item {inputs[0]} didn't contain any filepaths.")
 
-    absolute_path = str(Path(list(indexed_paths.values())[0]).resolve())
+    path = list(indexed_paths.values())[0]
+    if _is_remote_file(path):
+        return os.path.dirname(path)
+    absolute_path = str(Path(path).resolve())
 
-    if "/.project" in absolute_path:
-        return "/" + os.path.join(*str(list(indexed_paths.values())[0]).split("/")[:4])
-
-    return "/" + os.path.join(*str(absolute_path).split("/")[:4])
+    if _IS_IN_STUDIO or absolute_path.startswith("/teamspace"):
+        if "/.project" in absolute_path:
+            return "/" + os.path.join(*str(list(indexed_paths.values())[0]).split("/")[:4])
+        return "/" + os.path.join(*str(absolute_path).split("/")[:4])
+    return None
 
 
 def _get_default_num_workers() -> int:
@@ -78,7 +99,7 @@ def _get_default_num_workers() -> int:
 
 
 class LambdaDataTransformRecipe(DataTransformRecipe):
-    def __init__(self, fn: Callable[[str, Any], None], inputs: Sequence[Any]):
+    def __init__(self, fn: Callable[[str, Any], None], inputs: Union[Sequence[Any], StreamingDataLoader]):
         super().__init__()
         self._fn = fn
         self._inputs = inputs
@@ -124,40 +145,49 @@ class LambdaDataChunkRecipe(DataChunkRecipe):
     def __init__(
         self,
         fn: Callable[[Any], None],
-        inputs: Sequence[Any],
+        inputs: Union[Sequence[Any], StreamingDataLoader],
         chunk_size: Optional[int],
         chunk_bytes: Optional[Union[int, str]],
         compression: Optional[str],
+        encryption: Optional[Encryption] = None,
+        existing_index: Optional[Dict[str, Any]] = None,
     ):
-        super().__init__(chunk_size=chunk_size, chunk_bytes=chunk_bytes, compression=compression)
+        super().__init__(chunk_size=chunk_size, chunk_bytes=chunk_bytes, compression=compression, encryption=encryption)
         self._fn = fn
         self._inputs = inputs
+        self.is_generator = False
+        self.existing_index = existing_index
+
+        self.check_fn()
+
+        self.prepare_item = self._prepare_item_generator if self.is_generator else self._prepare_item  # type: ignore
+
+    def check_fn(self) -> None:
+        if (
+            isinstance(self._fn, (partial, FunctionType))
+            and inspect.isgeneratorfunction(self._fn)
+            or (callable(self._fn) and inspect.isgeneratorfunction(self._fn.__call__))  # type: ignore
+        ):
+            self.is_generator = True
+
+    def _prepare_item(self, item_metadata: Any) -> Any:
+        return self._fn(item_metadata)
+
+    def _prepare_item_generator(self, item_metadata: Any) -> Any:
+        yield from self._fn(item_metadata)  # type: ignore
 
     def prepare_structure(self, input_dir: Optional[str]) -> Any:
         return self._inputs
 
     def prepare_item(self, item_metadata: Any) -> Any:
-        if isinstance(self._fn, partial):
-            yield from self._fn(item_metadata)
-
-        elif isinstance(self._fn, FunctionType):
-            if inspect.isgeneratorfunction(self._fn):
-                yield from self._fn(item_metadata)
-            else:
-                yield self._fn(item_metadata)
-        elif callable(self._fn):
-            if inspect.isgeneratorfunction(self._fn.__call__):  # type: ignore
-                yield from self._fn.__call__(item_metadata)  # type: ignore
-            else:
-                yield self._fn.__call__(item_metadata)  # type: ignore
-        else:
-            raise ValueError(f"The provided {self._fn} isn't supported.")
+        """This method is overriden dynamically."""
 
 
 def map(
     fn: Callable[[str, Any], None],
-    inputs: Sequence[Any],
+    inputs: Union[Sequence[Any], StreamingDataLoader],
     output_dir: Union[str, Dir],
+    input_dir: Optional[str] = None,
     weights: Optional[List[int]] = None,
     num_workers: Optional[int] = None,
     fast_dev_run: Union[bool, int] = False,
@@ -170,13 +200,14 @@ def map(
     reader: Optional[BaseReader] = None,
     batch_size: Optional[int] = None,
 ) -> None:
-    """This function map a callbable over a collection of files possibly in a distributed way.
+    """This function maps a callable over a collection of inputs, possibly in a distributed way.
 
     Arguments:
         fn: A function to be executed over each input element
-        inputs: A sequence of input to be processed by the `fn` function.
-            Each input should contain at least a valid filepath.
+        inputs: A sequence of input to be processed by the `fn` function, or a streaming dataloader.
         output_dir: The folder where the processed data should be stored.
+        input_dir: Provide the path where your files are stored. If the files are on a remote storage,
+            they will be downloaded in the background while processed.
         weights: Provide an associated weight to each input. This is used to balance work among workers.
         num_workers: The number of workers to use during processing
         fast_dev_run: Whether to use process only a sub part of the inputs
@@ -197,7 +228,9 @@ def map(
         raise ValueError("When providing a streaming dataloader, weights isn't supported.")
 
     if not isinstance(inputs, (Sequence, StreamingDataLoader)):
-        raise ValueError(f"The provided inputs should be non empty sequence or a streaming dataloader. Found {inputs}.")
+        raise ValueError(
+            f"The provided inputs should be a non-empty sequence or a streaming dataloader. Found {inputs}."
+        )
 
     if len(inputs) == 0:
         raise ValueError(f"The provided inputs should be non empty. Found {inputs}.")
@@ -227,15 +260,19 @@ def map(
             _assert_dir_is_empty(_output_dir)
 
         if not isinstance(inputs, StreamingDataLoader):
-            input_dir = _resolve_dir(_get_input_dir(inputs))
+            input_dir = input_dir or _get_input_dir(inputs)
+            resolved_dir = _resolve_dir(input_dir)
 
             if isinstance(batch_size, int) and batch_size > 1:
                 inputs = [inputs[pos : pos + batch_size] for pos in range(0, len(inputs), batch_size)]
         else:
-            input_dir = Dir()
+            resolved_dir = Dir()
+
+        if num_workers == 0:
+            num_workers = 1
 
         data_processor = DataProcessor(
-            input_dir=input_dir,
+            input_dir=resolved_dir,
             output_dir=_output_dir,
             num_workers=num_workers or _get_default_num_workers(),
             fast_dev_run=fast_dev_run,
@@ -256,12 +293,14 @@ def map(
 
 def optimize(
     fn: Callable[[Any], Any],
-    inputs: Sequence[Any],
+    inputs: Union[Sequence[Any], StreamingDataLoader],
     output_dir: str,
+    input_dir: Optional[str] = None,
     weights: Optional[List[int]] = None,
     chunk_size: Optional[int] = None,
     chunk_bytes: Optional[Union[int, str]] = None,
     compression: Optional[str] = None,
+    encryption: Optional[Encryption] = None,
     num_workers: Optional[int] = None,
     fast_dev_run: bool = False,
     num_nodes: Optional[int] = None,
@@ -271,18 +310,26 @@ def optimize(
     reorder_files: bool = True,
     reader: Optional[BaseReader] = None,
     batch_size: Optional[int] = None,
+    mode: Optional[Literal["append", "overwrite"]] = None,
+    use_checkpoint: bool = False,
+    item_loader: Optional[BaseItemLoader] = None,
+    start_method: Optional[str] = None,
 ) -> None:
-    """This function converts a dataset into chunks possibly in a distributed way.
+    """This function converts a dataset into chunks, possibly in a distributed way.
 
     Arguments:
-        fn: A function to be executed over each input element
-        inputs: A sequence of input to be processed by the `fn` function.
-            Each input should contain at least a valid filepath.
+        fn: A function to be executed over each input element. The function should return the data sample that
+            corresponds to the input. Every invocation of the function should return a similar hierarchy of objects,
+            where the object types and list sizes don't change.
+        inputs: A sequence of input to be processed by the `fn` function, or a streaming dataloader.
         output_dir: The folder where the processed data should be stored.
+        input_dir: Provide the path where your files are stored. If the files are on a remote storage,
+            they will be downloaded in the background while processed.
         weights: Provide an associated weight to each input. This is used to balance work among workers.
         chunk_size: The maximum number of elements to hold within a chunk.
         chunk_bytes: The maximum number of bytes to hold within a chunk.
         compression: The compression algorithm to use over the chunks.
+        encryption: The encryption algorithm to use over the chunks.
         num_workers: The number of workers to use during processing
         fast_dev_run: Whether to use process only a sub part of the inputs
         num_nodes: When doing remote execution, the number of nodes to use. Only supported on https://lightning.ai/.
@@ -292,8 +339,19 @@ def optimize(
         reorder_files: By default, reorders the files by file size to distribute work equally among all workers.
             Set this to ``False`` if the order in which samples are processed should be preserved.
         batch_size: Group the inputs into batches of batch_size length.
+        mode: The mode to use when writing the data. Accepts either ``append`` or ``overwrite`` or None.
+            Defaults to None.
+        use_checkpoint: Whether to create checkpoints while processing the data, which can be used to resume the
+            processing from the last checkpoint if the process is interrupted. (`Default: False`)
+        item_loader: The item loader that will be used during loading in StreamingDataset. Determines
+                the format in which the data is stored and optimized for loading.
+        start_method: The start method used by python multiprocessing package. Default to spawn unless running
+            inside an interactive shell like Ipython.
 
     """
+    if mode is not None and mode not in ["append", "overwrite"]:
+        raise ValueError(f"The provided `mode` should be either `append` or `overwrite`. Found {mode}.")
+
     if isinstance(inputs, StreamingDataLoader) and batch_size is not None:
         raise ValueError("When providing a streaming dataloader, pass the batch_size to the dataloader directly.")
 
@@ -301,11 +359,12 @@ def optimize(
         raise ValueError("When providing a streaming dataloader, weights isn't supported.")
 
     if not isinstance(inputs, (Sequence, StreamingDataLoader)):
-        raise ValueError(f"The provided inputs should be non empty sequence or a streaming dataloader. Found {inputs}.")
+        raise ValueError(
+            f"The provided inputs should be a non-empty sequence or a streaming dataloader. Found {inputs}."
+        )
 
     if len(inputs) == 0:
         raise ValueError(f"The provided inputs should be non empty. Found {inputs}.")
-
     if chunk_size is None and chunk_bytes is None:
         raise ValueError("Either `chunk_size` or `chunk_bytes` needs to be defined.")
 
@@ -322,26 +381,53 @@ def optimize(
         )
 
     if num_nodes is None or int(os.getenv("DATA_OPTIMIZER_NUM_NODES", 0)) > 0:
+        DATA_OPTIMIZER_NUM_NODES = int(os.getenv("DATA_OPTIMIZER_NUM_NODES", 0))
         _output_dir: Dir = _resolve_dir(output_dir)
 
+        if (
+            _output_dir.url is None
+            and _output_dir.path
+            and _output_dir.path.startswith("/teamspace/studios/this_studio")
+            and DATA_OPTIMIZER_NUM_NODES > 0
+        ):
+            assert _output_dir.path
+            output_dir = _output_dir.path.replace("/teamspace/studios/this_studio", "")
+            output_dir = _get_work_dir().lstrip("/").rstrip("/") + "/" + output_dir.lstrip("/").rstrip("/")
+            _output_dir = _resolve_dir(output_dir)
+
         if _output_dir.url is not None and "cloudspaces" in _output_dir.url:
             raise ValueError(
                 f"The provided `output_dir` isn't valid. Found {_output_dir.path}."
                 " HINT: You can either use `/teamspace/s3_connections/...` or `/teamspace/datasets/...`."
             )
 
-        _assert_dir_has_index_file(_output_dir)
+        _assert_dir_has_index_file(_output_dir, mode=mode, use_checkpoint=use_checkpoint)
 
         if not isinstance(inputs, StreamingDataLoader):
-            input_dir = _resolve_dir(_get_input_dir(inputs))
+            resolved_dir = _resolve_dir(input_dir or _get_input_dir(inputs))
 
             if isinstance(batch_size, int) and batch_size > 1:
                 inputs = [inputs[pos : pos + batch_size] for pos in range(0, len(inputs), batch_size)]
         else:
-            input_dir = Dir()
+            resolved_dir = Dir()
+
+        if num_workers == 0:
+            num_workers = 1
+
+        num_workers = num_workers or _get_default_num_workers()
+        state_dict = {rank: 0 for rank in range(num_workers)}
+
+        existing_index_file_content = read_index_file_content(_output_dir) if mode == "append" else None
+
+        if existing_index_file_content is not None:
+            for chunk in existing_index_file_content["chunks"]:
+                rank, index = extract_rank_and_index_from_filename(chunk["filename"])
+
+                if rank < num_workers and state_dict[rank] <= index:
+                    state_dict[rank] = index + 1  # +1 because we want to start from the next index
 
         data_processor = DataProcessor(
-            input_dir=input_dir,
+            input_dir=resolved_dir,
             output_dir=_output_dir,
             num_workers=num_workers or _get_default_num_workers(),
             fast_dev_run=fast_dev_run,
@@ -349,6 +435,10 @@ def optimize(
             num_uploaders=num_uploaders,
             reorder_files=reorder_files,
             reader=reader,
+            state_dict=state_dict,
+            use_checkpoint=use_checkpoint,
+            item_loader=item_loader,
+            start_method=start_method,
         )
 
         with optimize_dns_context(True):
@@ -359,6 +449,8 @@ def optimize(
                     chunk_size=chunk_size,
                     chunk_bytes=chunk_bytes,
                     compression=compression,
+                    encryption=encryption,
+                    existing_index=existing_index_file_content,
                 )
             )
         return None
@@ -418,3 +510,120 @@ class walk:
                         future = executor.submit(_listdir, folder)
                         self.futures.append(future)
         return
+
+
+@dataclass
+class CopyInfo:
+    input_dir: Dir
+    old_filename: str
+    new_filename: str
+
+
+def merge_datasets(input_dirs: List[str], output_dir: str) -> None:
+    """The merge_datasets utility enables to merge multiple existing optimized datasets into a single optimized
+    dataset.
+
+    Arguments:
+        input_dirs: A list of directories pointing to the existing optimized datasets.
+        output_dir: The directory where the merged dataset would be stored.
+
+    """
+    if len(input_dirs) == 0:
+        raise ValueError("The input directories needs to be defined.")
+
+    if len(input_dirs) == 1:
+        raise ValueError("There should be more than 1 input directory")
+
+    resolved_input_dirs = [_resolve_dir(input_dir) for input_dir in input_dirs]
+    resolved_output_dir = _resolve_dir(output_dir)
+
+    if any(input_dir == resolved_output_dir for input_dir in resolved_input_dirs):
+        raise ValueError("The provided output_dir was found within the input_dirs. This isn't supported.")
+
+    input_dirs_file_content = [read_index_file_content(input_dir) for input_dir in resolved_input_dirs]
+
+    if any(file_content is None for file_content in input_dirs_file_content):
+        raise ValueError("One of the provided input_dir doesn't have an index file.")
+
+    output_dir_file_content = read_index_file_content(resolved_output_dir)
+
+    if output_dir_file_content is not None:
+        raise ValueError("The output_dir already contains an optimized dataset")
+
+    assert input_dirs_file_content
+
+    for input_dir_file_content in input_dirs_file_content[1:]:
+        if input_dirs_file_content[0]["config"]["data_format"] != input_dir_file_content["config"]["data_format"]:  # type: ignore
+            raise ValueError("Your are trying to merge datasets with different data formats")
+
+        if input_dirs_file_content[0]["config"]["compression"] != input_dir_file_content["config"]["compression"]:  # type: ignore
+            raise ValueError("Your are trying to merge datasets with different compression configuration.")
+
+    chunks = []
+    copy_infos: List[CopyInfo] = []
+    counter = 0
+    for input_dir, input_dir_file_content in zip(resolved_input_dirs, input_dirs_file_content):
+        compression = input_dir_file_content["config"]["compression"]  # type: ignore
+        for chunk in input_dir_file_content["chunks"]:  # type: ignore
+            assert isinstance(chunk, dict)
+            old_filename = chunk["filename"]
+            new_filename = (
+                f"chunk-0-{counter}.{compression}.bin" if compression is not None else f"chunk-0-{counter}.bin"
+            )
+            copy_infos.append(CopyInfo(input_dir=input_dir, old_filename=old_filename, new_filename=new_filename))
+            chunk["filename"] = new_filename
+            chunks.append(chunk)
+            counter += 1
+
+    index_json = {"config": input_dirs_file_content[0]["config"], "chunks": chunks}  # type: ignore
+
+    _tqdm = _get_tqdm_iterator_if_available()
+
+    for copy_info in _tqdm(copy_infos):
+        _apply_copy(copy_info, resolved_output_dir)
+
+    _save_index(index_json, resolved_output_dir)
+
+
+def _apply_copy(copy_info: CopyInfo, output_dir: Dir) -> None:
+    if output_dir.url is None and copy_info.input_dir.url is None:
+        assert copy_info.input_dir.path
+        assert output_dir.path
+        input_filepath = os.path.join(copy_info.input_dir.path, copy_info.old_filename)
+        output_filepath = os.path.join(output_dir.path, copy_info.new_filename)
+        os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
+        shutil.copyfile(input_filepath, output_filepath)
+
+    elif output_dir.url and copy_info.input_dir.url:
+        input_obj = parse.urlparse(os.path.join(copy_info.input_dir.url, copy_info.old_filename))
+        output_obj = parse.urlparse(os.path.join(output_dir.url, copy_info.new_filename))
+
+        s3 = S3Client()
+        s3.client.copy(
+            {"Bucket": input_obj.netloc, "Key": input_obj.path.lstrip("/")},
+            output_obj.netloc,
+            output_obj.path.lstrip("/"),
+        )
+    else:
+        raise NotImplementedError
+
+
+def _save_index(index_json: Dict, output_dir: Dir) -> None:
+    if output_dir.url is None:
+        assert output_dir.path
+        with open(os.path.join(output_dir.path, _INDEX_FILENAME), "w") as f:
+            json.dump(index_json, f)
+    else:
+        with tempfile.NamedTemporaryFile("w") as f:
+            json.dump(index_json, f)
+
+            f.flush()
+
+            obj = parse.urlparse(os.path.join(output_dir.url, _INDEX_FILENAME))
+
+            s3 = S3Client()
+            s3.client.upload_file(
+                f.name,
+                obj.netloc,
+                obj.path.lstrip("/"),
+            )
