@@ -33,7 +33,7 @@ from torch.utils.data.dataloader import (
 )
 from torch.utils.data.sampler import BatchSampler, Sampler
 
-from litdata.constants import _DEFAULT_CHUNK_BYTES, _TORCH_GREATER_EQUAL_2_1_0, _VIZ_TRACKER_AVAILABLE
+from litdata.constants import _DEFAULT_CHUNK_BYTES, _VIZ_TRACKER_AVAILABLE
 from litdata.streaming import Cache
 from litdata.streaming.combined import (
     __NUM_SAMPLES_YIELDED_KEY__,
@@ -42,11 +42,9 @@ from litdata.streaming.combined import (
 )
 from litdata.streaming.dataset import StreamingDataset
 from litdata.streaming.sampler import CacheBatchSampler
+from litdata.utilities._pytree import tree_flatten
 from litdata.utilities.env import _DistributedEnv
 
-if _TORCH_GREATER_EQUAL_2_1_0:
-    from torch.utils._pytree import tree_flatten
-
 logger = logging.Logger(__name__)
 
 
@@ -478,6 +476,21 @@ class _StreamingMultiProcessingDataLoaderIter(_MultiProcessingDataLoaderIter):
             super()._try_put_index()
 
 
+class StreamingDataLoaderCollateFn:
+    def __init__(self, collate_fn: Optional[Callable] = None) -> None:
+        self.collate_fn = collate_fn or default_collate
+
+    def __call__(self, items: List[Any]) -> Any:
+        if len(items) > 0 and isinstance(items[0], dict) and __NUM_SAMPLES_YIELDED_KEY__ in items[0]:
+            batch = self.collate_fn([item[__SAMPLES_KEY__] for item in items])
+            return {
+                __SAMPLES_KEY__: batch,
+                __NUM_SAMPLES_YIELDED_KEY__: list(torch.tensor(items[-1][__NUM_SAMPLES_YIELDED_KEY__]).reshape(-1, 1)),
+            }
+
+        return self.collate_fn(items)
+
+
 class StreamingDataLoader(DataLoader):
     r"""The StreamingDataLoader combines a dataset and a sampler, and provides an iterable over the given dataset.
 
@@ -542,6 +555,8 @@ class StreamingDataLoader(DataLoader):
         profile_dir: Optional[str] = None,
         prefetch_factor: Optional[int] = None,
         shuffle: Optional[bool] = None,
+        drop_last: Optional[bool] = None,
+        collate_fn: Optional[Callable] = None,
         **kwargs: Any,
     ) -> None:  # pyright: ignore
         if not isinstance(dataset, (StreamingDataset, CombinedStreamingDataset)):
@@ -553,6 +568,12 @@ class StreamingDataLoader(DataLoader):
         if shuffle is not None:
             dataset.set_shuffle(shuffle)
 
+        if drop_last is not None:
+            dataset.set_drop_last(drop_last)
+
+        dataset.set_batch_size(batch_size)
+        dataset.set_num_workers(num_workers)
+
         shuffle = None
 
         if profile_batches and not _VIZ_TRACKER_AVAILABLE:
@@ -561,6 +582,9 @@ class StreamingDataLoader(DataLoader):
         if profile_batches and num_workers == 0:
             raise ValueError("Profiling is supported only with num_workers >= 1.")
 
+        if collate_fn:
+            collate_fn = StreamingDataLoaderCollateFn(collate_fn)
+
         self.current_epoch = 0
         self.batch_size = batch_size
         self.num_workers = num_workers
@@ -579,6 +603,7 @@ class StreamingDataLoader(DataLoader):
             batch_size=batch_size,
             num_workers=num_workers,
             prefetch_factor=(10 if num_workers > 0 else None) if prefetch_factor is None else prefetch_factor,
+            collate_fn=collate_fn,
             **kwargs,
         )  # type: ignore
 
@@ -590,6 +615,7 @@ class StreamingDataLoader(DataLoader):
             self.current_epoch += 1
             self._num_samples_yielded_combined = {}
             self._num_samples_yielded_streaming = 0
+            self.dataset.reset_state_dict()
 
         self.dataset.set_epoch(self.current_epoch)
 
@@ -617,6 +643,16 @@ class StreamingDataLoader(DataLoader):
 
         self.restore = False
 
+    def __len__(self) -> int:
+        if self._dataset_kind == _DatasetKind.Iterable:
+            length = self._IterableDataset_len_called = self.dataset.get_len(self.num_workers, self.batch_size)
+            if self.batch_size is not None:  # IterableDataset doesn't allow custom sampler or batch_sampler
+                from math import ceil
+
+                return length // self.batch_size if self.drop_last else ceil(length / self.batch_size)
+            return length
+        return len(self._index_sampler)
+
     def state_dict(self) -> Dict[str, Any]:
         if isinstance(self.dataset, StreamingDataset):
             assert self.batch_size
@@ -665,13 +701,23 @@ class StreamingDataLoader(DataLoader):
 
         # Inform we are resuming and disable resetting the StreamingDataLoader state.
         # This is toggle back to False when the `__iter__` method of the StreamingDataLoader completes.
-        self.restore = True
+        # self.restore = True
 
         if isinstance(self.dataset, CombinedStreamingDataset):
             self.dataset._set_use_streaming_dataloader(True)
             self.dataset.load_state_dict(obj)
+
+            # Inform that the dataloader is resuming.
+            # TODO: Check if the number of samples yielded is less than the length of the dataset.
+            # Also, len is not available for CombinedStreamingDataset incase of provided weights.
+            self.restore = True
+
         elif isinstance(self.dataset, StreamingDataset):
             self.dataset.load_state_dict(obj["dataset"])
+
+            # Inform that the dataloader is resuming.
+            if self._num_samples_yielded_streaming < len(self.dataset):
+                self.restore = True
         else:
             raise RuntimeError("The provided dataset should be a `StreamingDataset` or a `CombinedStreamingDataset`.")
 
