@@ -11,7 +11,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import hashlib
 import os
 from logging import Logger
 from time import time
@@ -21,16 +20,22 @@ import numpy as np
 from torch.utils.data import IterableDataset
 
 from litdata.constants import (
-    _DEFAULT_CACHE_DIR,
     _INDEX_FILENAME,
 )
 from litdata.streaming import Cache
+from litdata.streaming.downloader import get_downloader_cls  # noqa: F401
 from litdata.streaming.item_loader import BaseItemLoader
 from litdata.streaming.resolver import Dir, _resolve_dir
 from litdata.streaming.sampler import ChunkedIndex
 from litdata.streaming.serializers import Serializer
 from litdata.streaming.shuffle import FullShuffle, NoShuffle, Shuffle
+from litdata.utilities.dataset_utilities import _should_replace_path, _try_create_cache_dir, subsample_streaming_dataset
+from litdata.utilities.encryption import Encryption
 from litdata.utilities.env import _DistributedEnv, _is_in_dataloader_worker, _WorkerEnv
+from litdata.utilities.shuffle import (
+    _find_chunks_per_workers_on_which_to_skip_deletion,
+    _map_node_worker_rank_to_chunk_indexes_to_not_delete,
+)
 
 logger = Logger(__name__)
 
@@ -47,6 +52,10 @@ class StreamingDataset(IterableDataset):
         seed: int = 42,
         serializers: Optional[Dict[str, Serializer]] = None,
         max_cache_size: Union[int, str] = "100GB",
+        subsample: float = 1.0,
+        encryption: Optional[Encryption] = None,
+        storage_options: Optional[Dict] = {},
+        max_pre_download: int = 2,
     ) -> None:
         """The streaming dataset can be used once your data have been optimised using the DatasetOptimiser class.
 
@@ -61,15 +70,27 @@ class StreamingDataset(IterableDataset):
             seed: Random seed for shuffling.
             serializers: The serializers used to serialize and deserialize the chunks.
             max_cache_size: The maximum cache size used by the StreamingDataset.
+            subsample: Float representing fraction of the dataset to be randomly sampled (e.g., 0.1 => 10% of dataset).
+            encryption: The encryption object to use for decrypting the data.
+            storage_options: Additional connection options for accessing storage services.
+            max_pre_download: Maximum number of chunks that can be pre-downloaded by the StreamingDataset.
 
         """
         super().__init__()
         if not isinstance(shuffle, bool):
             raise ValueError(f"Shuffle should be a boolean. Found {shuffle}")
 
+        if not isinstance(subsample, float) or subsample <= 0.0 or subsample > 1.0:
+            raise ValueError("subsample must be a float with value between 0 and 1.")
+
         input_dir = _resolve_dir(input_dir)
 
         self.input_dir = input_dir
+        self.subsampled_files: List[str] = []
+        self.region_of_interest: List[Tuple[int, int]] = []
+        self.subsampled_files, self.region_of_interest = subsample_streaming_dataset(
+            self.input_dir, item_loader, subsample, shuffle, seed, storage_options
+        )
 
         self.item_loader = item_loader
         self.shuffle: bool = shuffle
@@ -106,10 +127,20 @@ class StreamingDataset(IterableDataset):
         self.shuffler: Optional[Shuffle] = None
         self.serializers = serializers
         self._state_dict: Optional[Dict[str, Any]] = None
+        # Has slightly different meaning in the context of the dataset
+        # We consider `num_workers = 0` from `torch.utils.DataLoader` still as 1 worker (the main process)
+        self.num_workers: int = 1
+        self.batch_size: int = 1
+        self._encryption = encryption
+        self.storage_options = storage_options
+        self.max_pre_download = max_pre_download
 
     def set_shuffle(self, shuffle: bool) -> None:
         self.shuffle = shuffle
 
+    def set_drop_last(self, drop_last: bool) -> None:
+        self.drop_last = drop_last
+
     def set_epoch(self, current_epoch: int) -> None:
         """Set the current epoch to the dataset on epoch starts.
 
@@ -131,10 +162,15 @@ class StreamingDataset(IterableDataset):
 
         cache = Cache(
             input_dir=self.input_dir,
+            subsampled_files=self.subsampled_files,
+            region_of_interest=self.region_of_interest,
             item_loader=self.item_loader,
             chunk_bytes=1,
             serializers=self.serializers,
             max_cache_size=self.max_cache_size,
+            encryption=self._encryption,
+            storage_options=self.storage_options,
+            max_pre_download=self.max_pre_download,
         )
         cache._reader._try_load_config()
 
@@ -156,10 +192,22 @@ class StreamingDataset(IterableDataset):
         return FullShuffle(cache, seed, drop_last) if self.shuffle else NoShuffle(cache, seed, drop_last)
 
     def __len__(self) -> int:
+        return self.get_len(self.num_workers, self.batch_size if self.batch_size else 1)
+
+    def set_batch_size(self, batch_size: int) -> None:
+        self.batch_size = batch_size
+
+    def set_num_workers(self, num_workers: int) -> None:
+        self.num_workers = num_workers or 1
+
+    def get_len(self, num_workers: int, batch_size: int) -> int:
+        self.set_num_workers(num_workers)
+        self.set_batch_size(batch_size)
+        worker_env = _WorkerEnv.detect()
         if self.shuffler is None:
-            cache = self._create_cache(worker_env=_WorkerEnv.detect())
+            cache = self._create_cache(worker_env=worker_env)
             self.shuffler = self._create_shuffler(cache)
-        return self.shuffler.get_len(self.distributed_env, self.current_epoch)
+        return self.shuffler.get_len(self.distributed_env, self.num_workers, self.batch_size, self.current_epoch)
 
     def __iter__(self) -> "StreamingDataset":
         # When the StreamingDataset is used within map or optimize, let's refetch the distributed env.
@@ -176,35 +224,48 @@ class StreamingDataset(IterableDataset):
             state: Dict[str, Any] = self._state_dict
             self.current_epoch = state["current_epoch"]
 
-        chunks_per_replica, intervals_per_replica = self.shuffler.get_chunks_and_intervals_per_ranks(
-            self.distributed_env, self.current_epoch
+        workers_chunks, workers_intervals = self.shuffler.get_chunks_and_intervals_per_workers(
+            self.distributed_env, self.worker_env.world_size, self.batch_size, self.current_epoch
         )
-        chunks_replica = chunks_per_replica[self.distributed_env.global_rank % self.distributed_env.world_size]
-        intervals_replica = intervals_per_replica[self.distributed_env.global_rank % self.distributed_env.world_size]
+
+        worker_rank = self.distributed_env.global_rank * self.worker_env.world_size + self.worker_env.rank
+        self.worker_chunks = workers_chunks[worker_rank]
+        self.worker_intervals = workers_intervals[worker_rank]
+
+        # The max number of samples to return from `__next__` (in worker)
+        self.stop_length = sum(interval[2] - interval[1] for interval in self.worker_intervals)
 
         # Handle restart
         if self._state_dict:
-            self._resume(chunks_replica, intervals_replica)
+            self._resume(workers_chunks, workers_intervals)
         else:
-            chunks_per_replica, intervals_per_replica = self.shuffler.get_chunks_and_intervals_per_ranks(
-                self.distributed_env, self.current_epoch
+            # Find the chunks shared across all workers of the current node.
+            # For each shared chunk, find the rank and worker to use the chunk last and prevent
+            # premature deletion for the other workers.
+            node_size = self.distributed_env.world_size // self.distributed_env.num_nodes
+            first_rank_this_node = (self.distributed_env.global_rank // node_size) * node_size
+            num_workers_per_node = node_size * self.num_workers
+            worker_start = first_rank_this_node * num_workers_per_node
+            worker_end = worker_start + num_workers_per_node
+            local_rank = self.distributed_env.global_rank % node_size
+
+            chunks_indexes_skip_deletion = _find_chunks_per_workers_on_which_to_skip_deletion(
+                self.num_workers,
+                self.batch_size,
+                workers_chunks[worker_start:worker_end],
+                workers_intervals[worker_start:worker_end],
+            )
+            worker_node_rank_to_chunk_indexes = _map_node_worker_rank_to_chunk_indexes_to_not_delete(
+                chunks_indexes_skip_deletion
             )
-            chunks_replica = chunks_per_replica[self.distributed_env.global_rank % self.distributed_env.world_size]
-            intervals_replica = intervals_per_replica[
-                self.distributed_env.global_rank % self.distributed_env.world_size
-            ]
 
-            self.worker_chunks = []
-            self.worker_intervals = []
-
-            for i, (chunk_index, chunk_interval) in enumerate(zip(chunks_replica, intervals_replica)):
-                if i % self.worker_env.world_size != self.worker_env.rank:
-                    continue
-                self.worker_chunks.append(chunk_index)
-                self.worker_intervals.append(chunk_interval)
+            worker_rank_local_node = local_rank * self.num_workers + self.worker_env.rank
+            if worker_rank_local_node in worker_node_rank_to_chunk_indexes:
+                self.cache._reader.config.skip_chunk_indexes_deletion = worker_node_rank_to_chunk_indexes[
+                    worker_rank_local_node
+                ]
 
             self.num_chunks = len(self.worker_chunks)
-
             self.current_indexes = []
             self.chunk_index = 0
             self.global_index = 0
@@ -215,7 +276,7 @@ class StreamingDataset(IterableDataset):
 
         return self
 
-    def _resume(self, chunks_replica: List[int], intervals_replica: List[Any]) -> None:
+    def _resume(self, workers_chunks: List[List[int]], workers_intervals: List[Any]) -> None:
         assert self._state_dict
         assert self.worker_env
         assert self.shuffler
@@ -228,32 +289,37 @@ class StreamingDataset(IterableDataset):
         # TODO: Implement elastic sampling where the number of workers, ranks can change.
         num_samples_yielded = self._state_dict["num_samples_yielded"]
 
+        worker_start = self.distributed_env.global_rank * num_workers
+        worker_end = worker_start + num_workers
+
         # replay sampling from each worker / chunks using the batch size
-        workers_chunks, workers_intervals = _associate_chunks_to_workers(
-            num_workers, self.worker_env, chunks_replica, intervals_replica
-        )
         indexes = _replay_sampling(num_samples_yielded, batch_size, num_workers)
-        chunks_index, indexes = _replay_chunks_sampling(workers_intervals, indexes)
+        chunks_index, indexes = _replay_chunks_sampling(
+            workers_intervals={i: workers_intervals[j] for i, j in enumerate(range(worker_start, worker_end))},
+            indexes=indexes,
+        )
 
         # select the chunks and intervals associated to this worker
-        worker_rank = self.worker_env.rank
+        worker_rank = self.distributed_env.global_rank * self.worker_env.world_size + self.worker_env.rank
+        worker_local_rank = self.worker_env.rank
+
         self.num_chunks = len(workers_intervals[worker_rank])
-        self.chunk_index = chunks_index[worker_rank]
+        self.chunk_index = chunks_index[worker_local_rank]
         self.worker_chunks = workers_chunks[worker_rank]
         self.worker_intervals = workers_intervals[worker_rank]
 
         # replay the indexes for the current chunks
         interval = self.worker_intervals[self.chunk_index]
-        current_indexes = np.arange(interval[0], interval[1])
+        current_indexes = np.arange(interval[1], interval[2])
 
         # re-shuffle the indexes
         current_indexes = self.shuffler(current_indexes, self.num_chunks, self.current_epoch, self.chunk_index)
 
         # skip any indexes already consumed
-        current_indexes = current_indexes[indexes[worker_rank] :]
+        current_indexes = current_indexes[indexes[worker_local_rank] :]
         self.current_indexes = current_indexes
 
-        self.global_index = num_samples_yielded
+        self.global_index = indexes[worker_local_rank]
 
         # bump the chunk_index
         self.chunk_index += 1
@@ -264,26 +330,33 @@ class StreamingDataset(IterableDataset):
             self.cache = self._create_cache(worker_env=self.worker_env)
             self.shuffler = self._create_shuffler(self.cache)
         if isinstance(index, int):
-            index = ChunkedIndex(index, self.cache._get_chunk_index_from_index(index))
+            index = ChunkedIndex(*self.cache._get_chunk_index_from_index(index))
+        elif isinstance(index, slice):
+            start, stop, step = index.indices(len(self))
+            _my_indices = list(range(start, stop, step))
+            _my_cache_indices = [ChunkedIndex(*self.cache._get_chunk_index_from_index(idx)) for idx in _my_indices]
+            return [self.cache[chnk_idx] for chnk_idx in _my_cache_indices]
         return self.cache[index]
 
     def __next__(self) -> Any:
         # Prevent to create more batch on a given process
-        if self.global_index >= len(self):
+        if self.global_index >= self.stop_length:
             self.current_epoch += 1
+            self.reset_state_dict()
             raise StopIteration
 
         # Lazily re-populate the interval to reduce memory usage.
         if len(self.current_indexes) == 0:
             if self.chunk_index == self.num_chunks:
                 self.current_epoch += 1
+                self.reset_state_dict()
                 raise StopIteration
 
             # reset index
             self.index = 0
 
             interval = self.worker_intervals[self.chunk_index]
-            current_indexes = np.arange(interval[0], interval[1])
+            current_indexes = np.arange(interval[1], interval[2])
 
             assert self.shuffler is not None
             assert self.num_chunks is not None
@@ -300,7 +373,7 @@ class StreamingDataset(IterableDataset):
                 index=index,
                 chunk_index=self.worker_chunks[self.chunk_index - 1],
                 # We provide the chunks indexes only one the first
-                chunk_indexes=None if self.has_triggered_download else self.worker_chunks,
+                chunk_indexes=None if self.has_triggered_download else self.worker_chunks[self.chunk_index - 1 :],
                 is_last_index=(self.chunk_index - 1) == len(self.worker_intervals) and len(self.current_indexes) == 1,
             )
         )
@@ -321,7 +394,7 @@ class StreamingDataset(IterableDataset):
 
         return {
             "num_samples_yielded": num_samples_yielded,
-            "num_workers": num_workers,
+            "num_workers": num_workers or 1,
             "batch_size": batch_size,
             "current_epoch": self.current_epoch,
             "input_dir_path": self.input_dir.path,
@@ -331,6 +404,8 @@ class StreamingDataset(IterableDataset):
             "seed": self.seed,
             "world_size": self.distributed_env.world_size,
             "shuffle": self.shuffle,
+            "subsampled_files": self.subsampled_files,
+            "region_of_interest": self.region_of_interest,
         }
 
     def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
@@ -338,13 +413,15 @@ class StreamingDataset(IterableDataset):
             # the state is restored within the workers
             self._state_dict = state_dict
 
+    def reset_state_dict(self) -> None:
+        self._state_dict = None
+
     def _validate_state_dict(self) -> None:
         assert self._state_dict
         assert self.worker_env
         assert self.cache
 
         state: Dict[str, Any] = self._state_dict
-
         if state["shuffle"] != self.shuffle:
             raise ValueError(
                 "The provided `shuffle` state doesn't match the current one. "
@@ -398,24 +475,36 @@ class StreamingDataset(IterableDataset):
                 f"Found `{self.drop_last}` instead of `{state['drop_last']}`."
             )
 
+        if state["num_samples_yielded"] > len(self):
+            raise ValueError(
+                "The provided `num_samples_yielded` state is greater than the dataset length. "
+                f"Found `{state['num_samples_yielded']}` instead of `{len(self)}`."
+            )
 
-def _try_create_cache_dir(input_dir: Optional[str]) -> Optional[str]:
-    hash_object = hashlib.md5((input_dir or "").encode())  # noqa: S324
-    if "LIGHTNING_CLUSTER_ID" not in os.environ or "LIGHTNING_CLOUD_PROJECT_ID" not in os.environ:
-        cache_dir = os.path.join(_DEFAULT_CACHE_DIR, hash_object.hexdigest())
-        os.makedirs(cache_dir, exist_ok=True)
-        return cache_dir
-    cache_dir = os.path.join("/cache", "chunks", hash_object.hexdigest())
-    os.makedirs(cache_dir, exist_ok=True)
-    return cache_dir
+    def reset(self) -> None:
+        # undo all the properties associated with original dataset
+        default_properties: Dict[str, Any] = {
+            "cache": None,
+            "worker_env": None,
+            "worker_chunks": [],
+            "worker_intervals": [],
+            "current_indexes": [],
+            "chunk_index": 0,
+            "num_chunks": None,
+            "global_index": 0,
+            "index": 0,
+            "has_triggered_download": False,
+            "min_items_per_replica": None,
+            "current_epoch": 1,
+            "random_state": None,
+            "shuffler": None,
+            "_state_dict": None,
+            "num_workers": 1,
+            "batch_size": 1,
+        }
 
-
-def _should_replace_path(path: Optional[str]) -> bool:
-    """Whether the input path is a special path to be replaced."""
-    if path is None or path == "":
-        return True
-
-    return path.startswith("/teamspace/datasets/") or path.startswith("/teamspace/s3_connections/")
+        for prop, value in default_properties.items():
+            setattr(self, prop, value)
 
 
 def is_integer(value: str) -> bool:
@@ -426,28 +515,6 @@ def is_integer(value: str) -> bool:
         return False
 
 
-def _associate_chunks_to_workers(
-    num_workers: int, worker_env: _WorkerEnv, chunks_replica: List[int], intervals_replica: List[Any]
-) -> Any:
-    workers_chunks = {}
-    workers_intervals = {}
-
-    for worker_idx in range(num_workers):
-        worker_chunks = []
-        worker_intervals = []
-        for i, (chunk_index, chunk_interval) in enumerate(zip(chunks_replica, intervals_replica)):
-            if i % worker_env.world_size != worker_idx:
-                continue
-
-            worker_chunks.append(chunk_index)
-            worker_intervals.append(chunk_interval)
-
-        workers_chunks[worker_idx] = worker_chunks
-        workers_intervals[worker_idx] = worker_intervals
-
-    return workers_chunks, workers_intervals
-
-
 def _replay_sampling(num_samples_yielded: int, batch_size: int, num_workers: int) -> Dict[int, int]:
     """This function replays the sampling from the dataloader."""
     divisible_num_batches_yielded = num_samples_yielded // (num_workers * batch_size)
@@ -481,9 +548,12 @@ def _replay_chunks_sampling(
 
     for worker_idx, intervals in workers_intervals.items():
         for interval in intervals:
-            size = interval[-1] - interval[0]
+            size = interval[2] - interval[1]
             if indexes[worker_idx] >= size:
                 indexes[worker_idx] -= size
                 chunks_index[worker_idx] += 1
+            else:
+                # We've reached the chunk where resuming needs to take place (for this worker)
+                break
 
     return chunks_index, indexes
