@@ -13,24 +13,24 @@
 
 import json
 import os
+import uuid
 import warnings
 from dataclasses import dataclass
-from time import sleep
+from time import sleep, time
 from typing import Any, Dict, List, Optional, Tuple, Union
 
 import numpy as np
-import torch
 
-from litdata.constants import _INDEX_FILENAME, _TORCH_GREATER_EQUAL_2_1_0
+from litdata.constants import _INDEX_FILENAME
 from litdata.processing.utilities import get_worker_rank
 from litdata.streaming.compression import _COMPRESSORS, Compressor
+from litdata.streaming.item_loader import BaseItemLoader, PyTreeLoader
 from litdata.streaming.serializers import Serializer, _get_serializers
+from litdata.utilities._pytree import PyTree, tree_flatten, treespec_dumps
+from litdata.utilities.encryption import Encryption, EncryptionLevel
 from litdata.utilities.env import _DistributedEnv, _WorkerEnv
 from litdata.utilities.format import _convert_bytes_to_int, _human_readable_bytes
 
-if _TORCH_GREATER_EQUAL_2_1_0:
-    from torch.utils._pytree import PyTree, tree_flatten, treespec_dumps
-
 
 @dataclass
 class Item:
@@ -50,8 +50,11 @@ class BinaryWriter:
         chunk_size: Optional[int] = None,
         chunk_bytes: Optional[Union[int, str]] = None,
         compression: Optional[str] = None,
+        encryption: Optional[Encryption] = None,
         follow_tensor_dimension: bool = True,
         serializers: Optional[Dict[str, Serializer]] = None,
+        chunk_index: Optional[int] = None,
+        item_loader: Optional[BaseItemLoader] = None,
     ):
         """The BinaryWriter enables to chunk dataset into an efficient streaming format for cloud training.
 
@@ -60,10 +63,14 @@ class BinaryWriter:
             chunk_bytes: The maximum number of bytes within a chunk.
             chunk_size: The maximum number of items within a chunk.
             compression: The compression algorithm to use.
+            encryption: The encryption algorithm to use.
             serializers: Provide your own serializers.
+            chunk_index: The index of the chunk to start from.
 
         """
         self._cache_dir = cache_dir
+        if not os.path.exists(self._cache_dir):
+            os.makedirs(self._cache_dir, exist_ok=True)
 
         if (isinstance(self._cache_dir, str) and not os.path.exists(self._cache_dir)) or self._cache_dir is None:
             raise FileNotFoundError(f"The provided cache directory `{self._cache_dir}` doesn't exist.")
@@ -76,6 +83,8 @@ class BinaryWriter:
         self._chunk_size = chunk_size
         self._chunk_bytes = _convert_bytes_to_int(chunk_bytes) if isinstance(chunk_bytes, str) else chunk_bytes
         self._compression = compression
+        self._encryption = encryption
+        self._item_loader = item_loader or PyTreeLoader()
 
         self._data_format: Optional[List[str]] = None
         self._data_spec: Optional[PyTree] = None
@@ -91,7 +100,7 @@ class BinaryWriter:
             self._compressor: Compressor = _COMPRESSORS[self._compression]
 
         self._serialized_items: Dict[int, Item] = {}
-        self._chunk_index = 0
+        self._chunk_index = chunk_index or 0
         self._min_index: Optional[int] = None
         self._max_index: Optional[int] = None
         self._chunks_info: List[Dict[str, Any]] = []
@@ -101,6 +110,10 @@ class BinaryWriter:
         self._distributed_env = _DistributedEnv.detect()
         self._follow_tensor_dimension = follow_tensor_dimension
 
+        self._per_sample_num_bytes = 0
+        self._per_sample_num_items = 0
+        self.last_checkpoint_chunk_info: List[Dict[str, Any]] = []
+
     @property
     def filled(self) -> bool:
         """Returns whether the caching phase is done."""
@@ -136,6 +149,8 @@ class BinaryWriter:
             "chunk_bytes": self._chunk_bytes,
             "data_format": self._data_format,
             "data_spec": treespec_dumps(self._data_spec) if self._data_spec else None,
+            "encryption": self._encryption.state_dict() if self._encryption else None,
+            "item_loader": self._item_loader.__class__.__name__,
         }
 
     def serialize(self, items: Any) -> Tuple[bytes, Optional[int]]:
@@ -144,8 +159,6 @@ class BinaryWriter:
         # Flatten the items provided by the users
         flattened, data_spec = tree_flatten(items)
 
-        is_single_tensor = len(flattened) == 1 and isinstance(flattened[0], torch.Tensor)
-
         # Collect the sizes and associated bytes for each item
         sizes: List[int] = []
         data: List[bytes] = []
@@ -164,14 +177,7 @@ class BinaryWriter:
             # tiny optimization to avoid looping over all the data format
             self._serialize_with_data_format(flattened, sizes, data, self._data_format)
 
-        # If there is a single element and it is a tensor, enable continous array.
-        if is_single_tensor:
-            return data[0], flattened[0].shape[0]
-
-        # Concatenante into a single byte array
-        head = np.array(sizes, np.uint32).tobytes()
-        body = b"".join(data)
-        return head + body, None
+        return self._item_loader.encode_data(data, sizes, flattened)
 
     def _serialize(self, item: Any, sizes: List[int], data: List[bytes]) -> str:
         """Serialize a given item and append its size and bytes to the sizes and data array."""
@@ -229,6 +235,11 @@ class BinaryWriter:
 
         current_chunk_bytes = sum([item.bytes for item in items])
 
+        # Whether to encrypt the data at the chunk level
+        if self._encryption and self._encryption.level == EncryptionLevel.CHUNK:
+            data = self._encryption.encrypt(data)
+            current_chunk_bytes = len(data)
+
         if self._chunk_bytes and current_chunk_bytes > self._chunk_bytes:
             warnings.warn(
                 f"An item was larger than the target chunk size ({_human_readable_bytes(self._chunk_bytes)})."
@@ -277,35 +288,68 @@ class BinaryWriter:
         self.add_item(index, items)
 
     def add_item(self, index: int, items: Any) -> Optional[str]:
-        # Track the minimum index provided to the writer
-        # Serialize the items and store an Item object.
+        """Given an index and items will serialize the items and store an Item object to the growing
+        `_serialized_items`."""
+
         if index in self._serialized_items:
             raise ValueError(f"The provided index {index} already exists in the cache.")
 
         data, dim = self.serialize(items)
+
+        # Whether to encrypt the data at the sample level
+        if self._encryption and self._encryption.level == EncryptionLevel.SAMPLE:
+            data = self._encryption.encrypt(data)
+
         self._serialized_items[index] = Item(
             index=index,
             data=data,
             bytes=len(data),
             dim=dim,
         )
-
-        if not self._should_write():
+        if self._min_index is None:
+            # When processing the first item for the current chunk
+            indexes = list(self._serialized_items.keys())
+            self._max_index = self._min_index = indexes[0] if len(indexes) == 1 else min(*indexes)
+            self._per_sample_num_items = self._per_sample_num_bytes = 0
+            if not self._should_write():
+                return None
+        elif index < self._min_index:
+            # reset the "temp" chunk
+            self._max_index = self._min_index = index
+            self._per_sample_num_items = self._per_sample_num_bytes = 0
+            if not self._should_write():
+                return None
+        elif index == self._max_index:
+            if not self._should_write():
+                return None
+        else:
             return None
+
         filepath = os.path.join(self._cache_dir, self.get_chunk_filename())
+
         self.write_chunk()
+
+        # now to reset
         self._min_index = None
         self._max_index = None
+        self._per_sample_num_bytes = 0
+        self._per_sample_num_items = 0
+
         return filepath
 
     def _should_write(self) -> bool:
         # TODO: Misleading method name, it modifies `self._min_index` and `self._max_index`!
         if not self._serialized_items:
             return False
-        indexes = list(self._serialized_items.keys())
-        self._min_index = index = indexes[0] if len(indexes) == 1 else min(*indexes)
-        num_bytes = 0
-        num_items = 0
+
+        if not isinstance(self._max_index, int):
+            return False
+
+        # We have already validated the indexes from the interval `min_index` to `max_index`` are in `_serialized_items`
+        # Resetting the num_bytes and  num_items back the values.
+        num_bytes = self._per_sample_num_bytes
+        num_items = self._per_sample_num_items
+        index = self._max_index
         while True:
             item = self._serialized_items.get(index, None)
             if item:
@@ -318,6 +362,9 @@ class BinaryWriter:
                     self._max_index = index - 1
                     return True
             else:
+                self._per_sample_num_bytes = num_bytes
+                self._per_sample_num_items = num_items
+                self._max_index = index
                 return False
 
     def write_chunk_to_file(
@@ -392,14 +439,25 @@ class BinaryWriter:
 
         self._merge_no_wait(node_rank=node_rank)
 
-    def _merge_no_wait(self, node_rank: Optional[int] = None) -> None:
+    def _merge_no_wait(self, node_rank: Optional[int] = None, existing_index: Optional[Dict[str, Any]] = None) -> None:
         """Once all the workers have written their own index, the merge function is responsible to read and merge them
-        into a single index."""
+        into a single index.
+
+        Arguments:
+            node_rank: The node rank of the index file
+            existing_index: Existing index to be added to the newly created one.
+
+        """
         files = os.listdir(self._cache_dir)
         index_files = [f for f in files if f.endswith(_INDEX_FILENAME)]
 
         chunks_info = []
         config = None
+
+        if existing_index is not None:
+            chunks_info.extend(existing_index["chunks"])
+            config = existing_index["config"]
+
         for index_filename in sorted(index_files):
             chunk_path = os.path.join(self._cache_dir, index_filename)
             with open(chunk_path) as f:
@@ -411,7 +469,7 @@ class BinaryWriter:
                 elif config != data["config"]:
                     raise Exception(
                         "The config isn't consistent between chunks. This shouldn't have happened."
-                        f"Found {config} {data['config']}."
+                        f"Found {config}; {data['config']}."
                     )
 
                 chunks_info.extend(data["chunks"])
@@ -420,7 +478,8 @@ class BinaryWriter:
 
         if node_rank is None:
             with open(os.path.join(self._cache_dir, _INDEX_FILENAME), "w") as f:
-                json.dump({"chunks": chunks_info, "config": config}, f, sort_keys=True)
+                data = {"chunks": chunks_info, "config": config, "updated_at": str(time())}
+                json.dump(data, f, sort_keys=True)
         else:
             with open(os.path.join(self._cache_dir, f"{node_rank}-{_INDEX_FILENAME}"), "w") as f:
                 json.dump({"chunks": chunks_info, "config": config}, f, sort_keys=True)
@@ -447,3 +506,25 @@ class BinaryWriter:
                 data=b"",
             )
         return out
+
+    def save_checkpoint(self, checkpoint_dir: str = ".checkpoints") -> Optional[str]:
+        """Save the current state of the writer to a checkpoint."""
+        checkpoint_dir = os.path.join(self._cache_dir, checkpoint_dir)
+        if not os.path.exists(checkpoint_dir):
+            os.makedirs(checkpoint_dir, exist_ok=True)
+
+        if self._chunks_info == self.last_checkpoint_chunk_info:
+            # to avoid saving the same checkpoint twice
+            return None
+
+        unique_id = uuid.uuid4().hex
+        done_till_index = sum(chnk_info["chunk_size"] for chnk_info in self._chunks_info)
+
+        checkpoint_filepath = os.path.join(checkpoint_dir, f"checkpoint-{self.rank}-{unique_id}.json")
+
+        checkPoint = {"chunks": self._chunks_info, "config": self.get_config(), "done_till_index": done_till_index}
+
+        with open(checkpoint_filepath, "w") as f:
+            json.dump(checkPoint, f)
+
+        return checkpoint_filepath
